#!/private/tmp/venv/bin/python

import getpass
import netrc
import argparse
import os
import signal
import sys
import requests
import html2text

CERT_ENDPOINT = "/cred/proxyCert"
CERT_SERVER = "www.canfar.phys.uvic.ca"

"""

The code in this file is deprecated being replaced by certificate related code in the net.auth module.
It is kept here only to support users that still rely on getCert application

"""
def get_cert(cert_server=None, cert_filename=None,
             cert_endpoint=None, **kwargs):
    """Access the cadc certificate server.

    :param cert_filename: the name of the file to write the certificate to
    :param cert_server: the http server that will provide the certificate
    :ptype cert_server: str
    :param cert_endpoint: the endpoint on the server where the certificate service is
    :ptype cert_endpoint: str
    :param kwargs: not really any, but maybe daysValid.
    :ptype daysValid: int

    """

    cert_server = cert_server is None and CERT_SERVER or cert_server
    cert_endpoint = cert_endpoint is None and CERT_ENDPOINT or cert_endpoint

    if cert_filename is None:
        cert_filename = os.path.join(os.getenv("HOME", "/tmp"), ".ssl/cadcproxy.pem")

    dirname = os.path.dirname(cert_filename)
    try:
        os.makedirs(dirname)
    except OSError as oex:
        if os.path.isdir(dirname):
            pass
        elif oex.errno == 20 or oex.errno == 17:
            sys.stderr.write("%s : %s \n" % (str(oex), dirname))
            sys.stderr.write("Expected %s to be a directory.\n" % dirname)
            sys.exit(oex.errno)
        else:
            raise oex
    username, passwd = get_user_password(cert_server)

    url = "http://{0}/{1}".format(cert_server, cert_endpoint)
    resp = requests.get(url, params=kwargs, auth=(username, passwd))
    if resp.status_code != 200:
        raise OSError(resp.status_code, html2text.html2text(resp.content))
    with open(cert_filename, 'w') as w:
        w.write(resp.content)
    return resp


def get_user_password(realm):
    """"Getting the username/password for realm from .netrc filie.

    :param realm: the server realm this user/password combination is for
    :ptype realm: str
    """
    if os.access(os.path.join(os.environ.get('HOME', '/'), ".netrc"), os.R_OK):
        auth = netrc.netrc().authenticators(realm)
    else:
        auth = False
    if not auth:
        sys.stdout.write("{0} Username: ".format(realm))
        username = sys.stdin.readline().strip('\n')
        password = getpass.getpass().strip('\n')
    else:
        username = auth[0]
        password = auth[2]
    return username, password


def _main():

    def _signal_handler(signal, frame):
        sys.stderr.write("\n")
        sys.exit(-1)

    signal.signal(signal.SIGINT, _signal_handler)

    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description=("Retrieve a security certificate for interation with VOSpace. "
                                                  "Certificate will be valid for daysValid and stored as local file "
                                                  "cert_filename. First looks for an entry in the users .netrc "
                                                  "matching the realm {0}, the user is prompted for a username and "
                                                  "password if no entry is found.".format(CERT_SERVER)))

    parser.add_argument('--daysValid', type=int, default=10, help='Number of days the cetificate should be valid.')
    parser.add_argument('--cert-filename',
                        default=os.path.join(os.getenv('HOME', '/tmp'), '.ssl/cadcproxy.pem'),
                        help="Filesysm location to store the proxy certifcate.")
    parser.add_argument('--cert-server',
                        default=CERT_SERVER,
                        help="Certificate server network address.")

    args = parser.parse_args()

    retry = True
    while retry:
        try:
            get_cert(cert_server=args.cert_server,
                     daysValid=args.daysValid,
                     cert_filename=args.cert_filename)
            retry = False
        except OSError as ose:
            if ose.errno != 401:
                sys.stderr.write(str(ose))
                return getattr(ose, 'errno', 1)
            else:
                sys.stderr.write("Access denied\n")
        except Exception as ex:
            sys.stderr.write(str(ex))
            return getattr(ex, 'errno', 1)


if __name__ == '__main__':
    sys.exit(_main())



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
The cadcutils package contains libraries commonly use in CADC applications.

The main modules and packages in cadcutils are:

    -utils: collection of utilities for time formats, application logging
            and command line argument parsing
    -net: network utilities: authentication, Web service base client, etc
    -exceptions: exceptions thrown by the cadcutils libraries

"""

try:
    _PACKAGE_SETUP_
except NameError:
    # do any package imports here
    pass




# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2016.                            (c) 2016.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

import logging
import sys
import unittest

from argparse import ArgumentError
from six import StringIO
from six.moves.urllib.parse import urlparse
from mock import Mock, patch, MagicMock, ANY

from cadcutils.util import *

class MyExitError(Exception):
    pass

class UtilTests(unittest.TestCase):

    """ Class for testing cadc utilities """

    def test_ivoa_dates(self):
        """ Test the ivoa date formats functions date2ivoa and str2ivoa """

        expect_date = '2016-11-07T11:22:33.123'
        self.assertEqual(expect_date, date2ivoa(str2ivoa(expect_date)))

        with self.assertRaises(ValueError):
            str2ivoa('2011-01-01')

        # handling of None arguments
        self.assertEqual(None, date2ivoa(None))
        self.assertEqual(None, str2ivoa(None))
        
    def test_get_log_level(self):
        """ Test the handling of logging level control from command line arguments """
        
        parser = get_base_parser(subparsers=False)
        args = parser.parse_args(["--debug", "--resource-id", "www.some.resource"])
        self.assertEqual(logging.DEBUG, get_log_level(args))

        parser = get_base_parser(subparsers=False,
                                 default_resource_id='ivo://www.some.resource/resourceID')
        args = parser.parse_args(["--verbose"])
        self.assertEqual(logging.INFO, get_log_level(args))
        
        args = parser.parse_args(["--quiet"])
        self.assertEqual(logging.FATAL, get_log_level(args))
        
        args = parser.parse_args([])
        self.assertEqual(logging.ERROR, get_log_level(args))

        print ("passed log level tests")

    def test_init_logging_debug(self):
        """ Test the init_logging function """
        logger1 = get_logger('namspace1', log_level=logging.DEBUG)
        self.assertEqual(logging.DEBUG, logger1.getEffectiveLevel())
        logger2 = get_logger('namspace2')
        self.assertEqual(logging.ERROR, logger2.getEffectiveLevel())

    def test_shared_logger(self):
        """ Loggers with the same namespace share the
            same logging instances """
        logger1 = get_logger('namspace1', log_level=logging.DEBUG)
        self.assertEqual(logging.DEBUG, logger1.getEffectiveLevel())
        logger2 = get_logger('namspace1', log_level=logging.WARN)
        self.assertEqual(logging.WARN, logger1.getEffectiveLevel())
        self.assertEqual(logging.WARN, logger2.getEffectiveLevel())
        logger3 = get_logger('namspace2', log_level=logging.INFO)
        self.assertEqual(logging.INFO, logger3.getEffectiveLevel())
        self.assertEqual(logging.WARN, logger1.getEffectiveLevel())
        self.assertEqual(logging.WARN, logger2.getEffectiveLevel())

    def test_modify_log_level(self):
        logger = get_logger('test_modify_log_level', log_level=logging.INFO)
        self.assertEqual(logging.INFO, logger.getEffectiveLevel())
        logger = get_logger('test_modify_log_level', log_level=logging.DEBUG)
        self.assertEqual(logging.DEBUG, logger.getEffectiveLevel())
        logger.setLevel(logging.ERROR)
        self.assertEqual(logging.ERROR, logger.getEffectiveLevel())

    def test_modname_log_format(self):

        stdout_pointer = sys.stdout
        try:
            sys.stdout = StringIO()      # capture output
            logger = get_logger()
            logger.error('Test message')
            out = sys.stdout.getvalue()  # release output
            self.assertTrue('test_utils' in out)
        finally:
            sys.stdout.close()  # close the stream 
            sys.stdout = stdout_pointer  # restore original stdout

    def test_info_log_format(self):

        stdout_pointer = sys.stdout
        try:
            sys.stdout = StringIO()      # capture output
            logger = get_logger('test_info_log_format', log_level=logging.INFO)
            logger.info('Test message')
            out = sys.stdout.getvalue()  # release output
            self.assertTrue('INFO' in out)
            self.assertTrue('test_info_log_format' in out)
            self.assertTrue('Test message' in out)
        finally:
            sys.stdout.close()  # close the stream 
            sys.stdout = stdout_pointer  # restore original stdout

    def test_debug_log_format(self):

        stdout_pointer = sys.stdout
        try:
            sys.stdout = StringIO()      # capture output
            logger = get_logger('test_debug_log_format_namespace', log_level=logging.DEBUG)
            logger.debug('Test message')
            out = sys.stdout.getvalue()  # release output
            self.assertTrue('DEBUG' in out)
            self.assertTrue('test_debug_log_format_namespace' in out)
            self.assertTrue('test_utils.test_debug_log_format:' in out)
            self.assertTrue('Test message' in out)
        finally:
            sys.stdout.close()  # close the stream 
            sys.stdout = stdout_pointer  # restore original stdout

    @patch('sys.exit', Mock(side_effect=[ArgumentError(None, None),
                                         ArgumentError(None, None),
                                         ArgumentError(None, None),
                                         ArgumentError(None, None),
                                         ArgumentError(None, None)]))
    def test_base_parser(self):

        parser = get_base_parser(subparsers=False)
        resource_id = "ivo://www.some.resource/resourceid"
        args = parser.parse_args(["--resource-id", resource_id])
        self.assertEqual(urlparse(resource_id), args.resource_id)

        parser = get_base_parser(subparsers=False, default_resource_id=resource_id)
        args = parser.parse_args([])
        self.assertEqual(resource_id, args.resource_id)

        # missing resourceID
        parser = get_base_parser(subparsers=False)
        with self.assertRaises(ArgumentError):
            parser.parse_args([])

        # invalid resourceID (scheme)
        resource_id = "http://www.some.resource/resourceid"
        parser = get_base_parser(subparsers=False, default_resource_id=resource_id)
        with self.assertRaises(ArgumentError):
            args = parser.parse_args([])

        # invalid resourceID (missing resource id)
        resource_id = "ivo://www.some.resource"
        parser = get_base_parser(subparsers=False, default_resource_id=resource_id)
        with self.assertRaises(ArgumentError):
            args = parser.parse_args([])

        # create a base parser for subparsers but fail to add subparsers
        parser = get_base_parser()
        with self.assertRaises(RuntimeError):
            parser.parse_args(['-h'])

        # try to add a subparser to a base parser created for no subparsers
        parser = get_base_parser(subparsers=False)
        with self.assertRaises(RuntimeError):
            parser.add_subparsers(dest='cmd')

    @patch('sys.exit', Mock(side_effect=[MyExitError, MyExitError, MyExitError,
                                         MyExitError, MyExitError, MyExitError]))
    def test_base_parser_help(self):
        #help with a simple, no subparsers basic parser - these are the default arguments
        self.maxDiff = None
        expected_stdout = \
'''usage: cadc-client [-h] [--cert CERT | -n | --netrc-file NETRC_FILE | -u USER]
                   [--host HOST] --resource-id RESOURCE_ID [-d | -q | -v] [-V]

optional arguments:
  --cert CERT           location of your X509 certificate to use for
                        authentication (unencrypted, in PEM format)
  -d, --debug           debug messages
  -h, --help            show this help message and exit
  --host HOST           base hostname for services - used mainly for testing
                        (default: www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca)
  -n                    use .netrc in $HOME for authentication
  --netrc-file NETRC_FILE
                        netrc file to use for authentication
  -q, --quiet           run quietly
  --resource-id RESOURCE_ID
                        resource identifier (e.g. ivo://cadc.nrc.ca/service)
  -u, --user USER       name of user to authenticate. Note: application
                        prompts for the corresponding password!
  -v, --verbose         verbose messages
  -V, --version         show program's version number and exit
'''

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            with self.assertRaises(MyExitError):
                sys.argv = ["cadc-client", "--help"]
                parser = get_base_parser(subparsers=False, version=3.3)
                parser.parse_args()
            self.assertEqual(expected_stdout, stdout_mock.getvalue())
        #print(stdout_mock.getvalue())

        # same test but no version this time
        expected_stdout = \
'''usage: cadc-client [-h] [--cert CERT | -n | --netrc-file NETRC_FILE | -u USER]
                   [--host HOST] --resource-id RESOURCE_ID [-d | -q | -v]

optional arguments:
  --cert CERT           location of your X509 certificate to use for
                        authentication (unencrypted, in PEM format)
  -d, --debug           debug messages
  -h, --help            show this help message and exit
  --host HOST           base hostname for services - used mainly for testing
                        (default: www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca)
  -n                    use .netrc in $HOME for authentication
  --netrc-file NETRC_FILE
                        netrc file to use for authentication
  -q, --quiet           run quietly
  --resource-id RESOURCE_ID
                        resource identifier (e.g. ivo://cadc.nrc.ca/service)
  -u, --user USER       name of user to authenticate. Note: application
                        prompts for the corresponding password!
  -v, --verbose         verbose messages
'''

        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            with self.assertRaises(MyExitError):
                sys.argv = ["cadc-client", "--help"]
                parser = get_base_parser(subparsers=False)
                parser.parse_args()
            self.assertEqual(expected_stdout, stdout_mock.getvalue())

        # --help with a simple parser with a few extra command line options
        expected_stdout = \
'''usage: cadc-client [-h] [--cert CERT | -n | --netrc-file NETRC_FILE | -u USER]
                   [--host HOST] --resource-id RESOURCE_ID [-d | -q | -v] [-x]
                   fileID [fileID ...]

positional arguments:
  fileID                the ID of the file in the archive

optional arguments:
  --cert CERT           location of your X509 certificate to use for
                        authentication (unencrypted, in PEM format)
  -d, --debug           debug messages
  -h, --help            show this help message and exit
  --host HOST           base hostname for services - used mainly for testing
                        (default: www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca)
  -n                    use .netrc in $HOME for authentication
  --netrc-file NETRC_FILE
                        netrc file to use for authentication
  -q, --quiet           run quietly
  --resource-id RESOURCE_ID
                        resource identifier (e.g. ivo://cadc.nrc.ca/service)
  -u, --user USER       name of user to authenticate. Note: application
                        prompts for the corresponding password!
  -v, --verbose         verbose messages
  -x                    test argument
'''
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            with self.assertRaises(MyExitError):
                sys.argv = ["cadc-client", "--help"]
                parser = get_base_parser(subparsers=False)
                parser.add_argument('-x', action='store_true', help='test argument')
                parser.add_argument('fileID',
                             help='the ID of the file in the archive', nargs='+')
                parser.parse_args()
            self.assertEqual(expected_stdout, stdout_mock.getvalue())


        #help with a parser with 2 subcommands
        parser = get_base_parser()
        subparsers = parser.add_subparsers(dest='cmd', help='My subcommands')
        parser_cmd1 = subparsers.add_parser('cmd1')
        parser_cmd1.add_argument('-x', action='store_true', help='test argument')
        parser_cmd2 = subparsers.add_parser('cmd2')
        parser_cmd2.add_argument('fileID',
                                 help='the ID of the file in the archive', nargs='+')

        expected_stdout = \
'''usage: cadc-client [-h] {cmd1,cmd2} ...

positional arguments:
  {cmd1,cmd2}  My subcommands

optional arguments:
  -h, --help   show this help message and exit
'''
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            with self.assertRaises(MyExitError):
                sys.argv = ["cadc-client", "-h"]
                parser.parse_args()
            self.assertEqual(expected_stdout, stdout_mock.getvalue())

        expected_stdout = \
'''usage: cadc-client cmd1 [-h]
                        [--cert CERT | -n | --netrc-file NETRC_FILE | -u USER]
                        [--host HOST] --resource-id RESOURCE_ID [-d | -q | -v]
                        [-x]

optional arguments:
  --cert CERT           location of your X509 certificate to use for
                        authentication (unencrypted, in PEM format)
  -d, --debug           debug messages
  -h, --help            show this help message and exit
  --host HOST           base hostname for services - used mainly for testing
                        (default: www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca)
  -n                    use .netrc in $HOME for authentication
  --netrc-file NETRC_FILE
                        netrc file to use for authentication
  -q, --quiet           run quietly
  --resource-id RESOURCE_ID
                        resource identifier (e.g. ivo://cadc.nrc.ca/service)
  -u, --user USER       name of user to authenticate. Note: application
                        prompts for the corresponding password!
  -v, --verbose         verbose messages
  -x                    test argument
'''
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            with self.assertRaises(MyExitError):
                sys.argv = ['cadc-client', 'cmd1', '-h']
                parser.parse_args()
            self.assertEqual(expected_stdout, stdout_mock.getvalue())

        expected_stdout = \
'''usage: cadc-client cmd2 [-h]
                        [--cert CERT | -n | --netrc-file NETRC_FILE | -u USER]
                        [--host HOST] --resource-id RESOURCE_ID [-d | -q | -v]
                        fileID [fileID ...]

positional arguments:
  fileID                the ID of the file in the archive

optional arguments:
  --cert CERT           location of your X509 certificate to use for
                        authentication (unencrypted, in PEM format)
  -d, --debug           debug messages
  -h, --help            show this help message and exit
  --host HOST           base hostname for services - used mainly for testing
                        (default: www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca)
  -n                    use .netrc in $HOME for authentication
  --netrc-file NETRC_FILE
                        netrc file to use for authentication
  -q, --quiet           run quietly
  --resource-id RESOURCE_ID
                        resource identifier (e.g. ivo://cadc.nrc.ca/service)
  -u, --user USER       name of user to authenticate. Note: application
                        prompts for the corresponding password!
  -v, --verbose         verbose messages
'''
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            with self.assertRaises(MyExitError):
                sys.argv = ['cadc-client', 'cmd2', '-h']
                parser.parse_args()
            self.assertEqual(expected_stdout, stdout_mock.getvalue())


# -*- coding: utf-8 -*-
# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2016.                            (c) 2016.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import unittest

import os
import requests
import time
from mock import Mock, patch, call, mock_open
from six.moves.urllib.parse import urlparse
from six import StringIO

from cadcutils.net import ws, auth
from cadcutils.net.ws import DEFAULT_RETRY_DELAY, MAX_RETRY_DELAY, MAX_NUM_RETRIES, SERVICE_RETRY
from cadcutils import exceptions

# The following is a temporary workaround for Python issue 25532 (https://bugs.python.org/issue25532)
call.__wrapped__ = None


class TestListResources(unittest.TestCase):

    @patch('cadcutils.net.ws.requests.get')
    def test_list_resources(self, get_mock):
        response_caps = Mock()
        response_caps.text = ('# This is just a test\n'
                                 'ivo://cadc.nrc.ca/serv1 = http://www.cadc.nrc.gc.ca/serv1/capabilities\n'
                                 'ivo://cadc.nrc.ca/serv2 = http://www.cadc.nrc.gc.ca/serv2/capabilities\n')
        response_serv1 = Mock()
        response_serv1.text = ('<vosi:capabilities xmlns:vosi="http://www.ivoa.net/xml/VOSICapabilities/v1.0" '
                                  'xmlns:vs="http://www.ivoa.net/xml/VODataService/v1.1" '
                                  'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\n'
                                  '<capability standardID="ivo://ivoa.net/std/VOSI#capabilities">\n'
                                  '<interface xsi:type="vs:ParamHTTP" role="std">\n'
                                  '<accessURL use="full">http://www.cadc.hia.nrcgc..ca/serv1/capabilities</accessURL>\n'
                                  '</interface>\n'
                                  '</capability>\n'
                                  '<capability standardID="ivo://ivoa.net/std/VOSI#availability">\n'
                                  '<interface xsi:type="vs:ParamHTTP" role="std">\n'
                                  '<accessURL use="full">http://www.cadc.nrc.gc.ca/serv1/availability</accessURL>\n'
                                  '</interface>\n'
                                  '</capability>\n'
                                  '</vosi:capabilities>\n')
        response_serv2 = Mock()
        response_serv2.text = ('<vosi:capabilities xmlns:vosi="http://www.ivoa.net/xml/VOSICapabilities/v1.0" '
                                  'xmlns:vs="http://www.ivoa.net/xml/VODataService/v1.1" '
                                  'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">\n'
                                 '<capability standardID="ivo://ivoa.net/std/VOSI#capabilities">\n'
                                 '<interface xsi:type="vs:ParamHTTP" role="std">\n'
                                 '<accessURL use="full">http://www.cadc.hia.nrcgc..ca/serv2/capabilities</accessURL>\n'
                                 '</interface>\n'
                                 '</capability>\n'
                                 '<capability standardID="ivo://ivoa.net/std/VOSI#tables-1.1">\n'
                                 '<interface xsi:type="vs:ParamHTTP" role="std">\n'
                                 '<accessURL use="full">http://www.cadc.nrc.gc.ca/serv2/availability</accessURL>\n'
                                 '</interface>\n'
                                 '</capability>\n'
                                 '</vosi:capabilities>\n')
        get_mock.side_effect = [response_caps, response_serv1, response_serv2]
        self.maxDiff = None
        usage = \
'''ivo://cadc.nrc.ca/serv1 (http://www.cadc.nrc.gc.ca/serv1/capabilities) - Capabilities: ivo://ivoa.net/std/VOSI#availability, ivo://ivoa.net/std/VOSI#capabilities

ivo://cadc.nrc.ca/serv2 (http://www.cadc.nrc.gc.ca/serv2/capabilities) - Capabilities: ivo://ivoa.net/std/VOSI#capabilities, ivo://ivoa.net/std/VOSI#tables-1.1

'''
        with patch('sys.stdout', new_callable=StringIO) as stdout_mock:
            ws.list_resources()
            self.assertEqual(usage, stdout_mock.getvalue())


class TestWs(unittest.TestCase):

    """Class for testing the webservie client"""
    @patch('cadcutils.net.ws.WsCapabilities')
    @patch('cadcutils.net.auth.os.path.isfile', Mock())
    @patch('cadcutils.net.auth.netrclib')
    @patch('cadcutils.net.ws.RetrySession.put')
    @patch('cadcutils.net.ws.RetrySession.head')
    @patch('cadcutils.net.ws.RetrySession.delete')
    @patch('cadcutils.net.ws.RetrySession.get')
    @patch('cadcutils.net.ws.RetrySession.post')
    def test_ops(self, post_mock, get_mock, delete_mock, head_mock, put_mock, netrclib_mock, caps_mock):
        anon_subject = auth.Subject()
        with self.assertRaises(ValueError):
            ws.BaseWsClient(None, anon_subject, "TestApp")
        resource = 'aresource'
        service = 'myservice'
        resource_id = 'ivo://www.canfar.phys.uvic.ca/{}'.format(service)
        # test anonymous access
        caps_mock.get_service_host.return_value = 'http://host/myservice'
        client = ws.BaseWsClient(resource_id, anon_subject, 'TestApp')
        resource_uri = urlparse(resource_id)
        base_url = 'http://{}{}/pub'.format(resource_uri.netloc, resource_uri.path)
        resource_url = 'http://{}{}/{}'.format(resource_uri.netloc, base_url, resource)
        self.assertEqual(anon_subject, client.subject)
        self.assertTrue(client.retry)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        self.assertEqual(None, client._session)  # lazy initialization
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test basic authentication access
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        host = 'www.different.org'
        subject = auth.Subject(netrc='somecert')
        client = ws.BaseWsClient(resource_id, subject, 'TestApp', retry=False, host=host)
        base_url = 'http://{}{}/auth'.format(host, resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertFalse(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test cert authentication
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        certfile = 'some/certfile.pem'
        subject = auth.Subject(certificate=certfile)
        client = ws.BaseWsClient(resource_id, subject, 'TestApp')
        base_url = 'https://{}{}/pub'.format(resource_uri.netloc, resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))
        self.assertEqual((certfile, certfile), client._session.cert)

        # repeat above tests test with the temporary sc2repo test
        resource = 'aresource'
        service = 'sc2repo'
        resource_id = 'ivo://www.canfar.phys.uvic.ca/{}'.format(service)
        # test anonymous access
        client = ws.BaseWsClient(resource_id, auth.Subject(), 'TestApp')
        resource_uri = urlparse(resource_id)
        base_url = 'http://{}{}/observations'.format(resource_uri.netloc, resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertTrue(client.retry)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        self.assertEqual(None, client._session)  # lazy initialization
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test basic authentication access
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        host = 'caom2workshop.canfar.net'
        subject = auth.Subject(netrc=True)
        subject._hosts_auth[host] = ('auser', 'apasswd')
        client = ws.BaseWsClient(resource_id, subject, 'TestApp', retry=False, host=host)
        base_url = 'http://{}{}/auth-observations'.format(host, resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertFalse(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))

        # test cert authentication
        post_mock.reset_mock()
        get_mock.reset_mock()
        put_mock.reset_mock()
        delete_mock.reset_mock()
        head_mock.reset_mock()
        certfile = 'some/certfile.pem'
        client = ws.BaseWsClient(resource_id, auth.Subject(certificate=certfile), 'TestApp')
        base_url = 'https://{}{}/observations'.format(resource_uri.netloc, resource_uri.path)
        resource_url = '{}/{}'.format(base_url, resource)
        self.assertEqual('TestApp', client.agent)
        self.assertTrue(client.retry)
        client.get(resource_url)
        get_mock.assert_called_with(resource_url, params=None)
        params = {'arg1': 'abc', 'arg2': 123, 'arg3': True}
        client.post(resource_url, **params)
        post_mock.assert_called_with(resource_url, **params)
        client.delete(resource_url)
        delete_mock.assert_called_with(resource_url)
        client.head(resource_url)
        head_mock.assert_called_with(resource_url)
        client.put(resource_url, **params)
        put_mock.assert_called_with(resource_url, **params)
        self.assertTrue(isinstance(client._session, ws.RetrySession))
        self.assertEqual((certfile, certfile), client._session.cert)

class TestRetrySession(unittest.TestCase):

    """ Class for testing retry session """

    @patch('time.sleep')
    @patch('cadcutils.net.ws.requests.Session.send')
    @patch('cadcutils.net.ws.requests.Session.merge_environment_settings', Mock(return_value={}))
    def test_retry(self, send_mock, time_mock):
        request = Mock()
        send_mock.return_value = Mock()
        rs = ws.RetrySession(False)
        rs.send(request)
        send_mock.assert_called_with(request)

        # mock delays for the 'Connection reset by peer error'
        # one connection error delay = DEFAULT_RETRY_DELAY
        send_mock.reset_mock()
        rs = ws.RetrySession()
        ce = requests.exceptions.ConnectionError()  # connection error that triggers retries
        ce.errno = 104
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [ce, response]
        rs.send(request)
        time_mock.assert_called_with(DEFAULT_RETRY_DELAY)

        # two connection error delay = DEFAULT_RETRY_DELAY
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession()
        ce = requests.exceptions.ConnectionError()  # connection error that triggers retries
        ce.errno = 104
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [ce, ce, response]  # two connection errors
        rs.send(request)
        calls = [call(DEFAULT_RETRY_DELAY), call(DEFAULT_RETRY_DELAY*2)]
        time_mock.assert_has_calls(calls)

        # set the start retry to a large number and see how it is capped to MAX_RETRY_DELAY
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession(start_delay=MAX_RETRY_DELAY/2 + 1)
        ce = requests.exceptions.ConnectionError()  # connection error that triggers retries
        ce.errno = 104
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [ce, ce, response]  # two connection errors
        rs.send(request)
        calls = (call(MAX_RETRY_DELAY/2 + 1), call(MAX_RETRY_DELAY))
        time_mock.assert_has_calls(calls)

        # return the error all the time
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession(start_delay=MAX_RETRY_DELAY/2 + 1)
        ce = requests.exceptions.ConnectionError()  # connection error that triggers retries
        ce.errno = 104
        # make sure the mock returns more errors than the maximum number of retries allowed
        http_errors = []
        i = 0
        while i <= MAX_NUM_RETRIES:
            http_errors.append(ce)
            i += 1
        send_mock.side_effect = http_errors
        with self.assertRaises(exceptions.HttpException):
            rs.send(request)

        # return the connection error other than 104 - connection reset by peer
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession(start_delay=MAX_RETRY_DELAY / 2 + 1)
        ce = requests.exceptions.ConnectionError()  # connection error that triggers retries
        ce.errno = 105
        send_mock.side_effect = ce
        with self.assertRaises(exceptions.HttpException):
            rs.send(request)

        # return HttpError 503 with Retry-After
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession()
        server_delay = 5
        he = requests.exceptions.HTTPError()  # connection error that triggers retries
        he.response = requests.Response()
        he.response.status_code = requests.codes.unavailable
        he.response.headers[SERVICE_RETRY] = server_delay
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [he, response]
        rs.send(request)
        calls = [call(server_delay)]
        time_mock.assert_has_calls(calls)

        # return HttpError 503 with Retry-After with an invalid value
        send_mock.reset_mock()
        time_mock.reset_mock()
        start_delay = 66
        rs = ws.RetrySession(start_delay=start_delay)
        server_delay = 'notnumber'
        he = requests.exceptions.HTTPError()  # connection error that triggers retries
        he.response = requests.Response()
        he.response.status_code = requests.codes.unavailable
        he.response.headers[SERVICE_RETRY] = server_delay
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [he, response]
        rs.send(request)
        calls = [call(start_delay)]  # uses the default delay
        time_mock.assert_has_calls(calls)

        # return HttpError 503 with no Retry-After
        send_mock.reset_mock()
        time_mock.reset_mock()
        start_delay = 66
        rs = ws.RetrySession(start_delay=start_delay)
        he = requests.exceptions.HTTPError()  # connection error that triggers retries
        he.response = requests.Response()
        he.response.status_code = requests.codes.unavailable
        response = requests.Response()
        response.status_code = requests.codes.ok
        send_mock.side_effect = [he, response]
        rs.send(request)
        calls = [call(start_delay)]  # uses the default delay
        time_mock.assert_has_calls(calls)

        # tests non-transient errors
        send_mock.reset_mock()
        time_mock.reset_mock()
        rs = ws.RetrySession()
        he = requests.exceptions.HTTPError()  # connection error that triggers retries
        he.response = requests.Response()
        he.response.status_code = requests.codes.internal_server_error
        send_mock.side_effect = he
        with self.assertRaises(exceptions.HttpException):
            rs.send(request)


    def test_config_file_location(self):
        # test the location of the config file when the host is specified
        myhost = 'myhost.ca'
        resource = 'aresource'
        service = 'myservice'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        client = Mock(resource_id=resource_id)
        with patch('cadcutils.net.ws.os.makedirs') as makedirs_mock:
            wscap = ws.WsCapabilities(client, host=myhost)
        expected_location = os.path.join(ws.CACHE_LOCATION, 'alt-domains', myhost)
        makedirs_mock.check_called_with(expected_location)
        self.assertEqual(os.path.join(expected_location, 'resource-caps'), wscap.reg_file)
        self.assertEqual(os.path.join(expected_location, urlparse(resource_id).netloc, service),
                         wscap.caps_file)

    def test_misc(self):
        """
        Tests miscellaneous functions
        """
        service_url = 'http://somehost.com/service'
        with patch('cadcutils.net.ws.WsCapabilities') as caps_mock:
            caps_mock.return_value.get_service_host.return_value = 'somehost.com'
            caps_mock.return_value.get_access_url.return_value = service_url
            client = ws.BaseWsClient("someresourceID", auth.Subject(), 'TestApp')
            self.assertEqual('{}'.format(service_url), client._get_url(('myfeature', None)))
            caps_mock.return_value.get_access_url.assert_called_once_with('myfeature')
            self.assertEqual('{}'.format(service_url), client._get_url(('myfeature', '')))
            
            
        test_host = 'testhost.com'
        with patch('cadcutils.net.ws.os.makedirs'):
            client = ws.BaseWsClient("someresourceID", auth.Subject(), 'TestApp', host=test_host)
        self.assertEqual(test_host, client.host)

        # test with resource as url
        with patch('cadcutils.net.ws.WsCapabilities') as caps_mock:
            caps_mock.return_value.get_service_host.return_value = 'somehost.com'
            client = ws.BaseWsClient("someresourceID", auth.Subject(), 'TestApp')
            resource_url = 'http://someurl.com/path/'
            self.assertEqual(resource_url, client._get_url(resource_url))
            # repeat with overriden host name
            client = ws.BaseWsClient("someresourceID", auth.Subject(), 'TestApp', host=test_host)
            # same result
            self.assertEqual(resource_url, client._get_url(resource_url))

            # test exceptions with different status in the response
            session = ws.RetrySession()
            response = requests.Response()
            response.status_code = requests.codes.not_found
            with self.assertRaises(exceptions.NotFoundException):
                session.check_status(response)
            response.status_code = requests.codes.unauthorized
            with self.assertRaises(exceptions.UnauthorizedException):
                session.check_status(response)
            response.status_code = requests.codes.forbidden
            with self.assertRaises(exceptions.ForbiddenException):
                session.check_status(response)
            response.status_code = requests.codes.bad_request
            with self.assertRaises(exceptions.BadRequestException):
                session.check_status(response)
            response.status_code = requests.codes.conflict
            with self.assertRaises(exceptions.AlreadyExistsException):
                session.check_status(response)
            response.status_code = requests.codes.internal_server_error
            with self.assertRaises(exceptions.InternalServerException):
                session.check_status(response)
            response.status_code = requests.codes.unavailable
            with self.assertRaises(requests.HTTPError):
                session.check_status(response)
            response.status_code = requests.codes.not_extended
            with self.assertRaises(exceptions.UnexpectedException):
                session.check_status(response)



capabilities__content = \
"""
<vosi:capabilities xmlns:vosi="http://www.ivoa.net/xml/VOSICapabilities/v1.0" xmlns:vs="http://www.ivoa.net/xml/VODataService/v1.1" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
    <capability standardID="ivo://ivoa.net/std/VOSI#capabilities">
        <interface xsi:type="vs:ParamHTTP" role="std">
            <accessURL use="full">
                http://WS_URL/capabilities
            </accessURL>
        </interface>
    </capability>
    <capability standardID="ivo://ivoa.net/std/VOSI#availability">
        <interface xsi:type="vs:ParamHTTP" role="std">
            <accessURL use="full">
                http://WS_URL/availability
            </accessURL>
        </interface>
    </capability>
    <capability standardID="vos://cadc.nrc.ca~service/CADC/mystnd01">
        <interface xsi:type="vs:ParamHTTP" role="std" version="1.0">
            <accessURL use="base">
                http://WS_URL/pub
            </accessURL>
        </interface>
        <interface xsi:type="vs:ParamHTTP" role="std" version="1.0">
            <accessURL use="base">
                http://WS_URL/auth
            </accessURL>
            <securityMethod standardID="http://www.w3.org/Protocols/HTTP/1.0/spec.html#BasicAA"/>
        </interface>
        <interface xsi:type="vs:ParamHTTP" role="std" version="1.0">
            <accessURL use="base">
                https://WS_URL
            </accessURL>
            <securityMethod standardID="ivo://ivoa.net/sso#tls-with-certificate"/>
        </interface>
    </capability>
</vosi:capabilities>
"""

class TestWsCapabilities(unittest.TestCase):

    """Class for testing the webservie client"""
    @patch('cadcutils.net.ws.os.path.getmtime')
    @patch('cadcutils.net.ws.open', mock=mock_open())
    @patch('cadcutils.net.ws.BaseWsClient.get')
    def test_get_reg(self, get_mock, file_mock, file_modtime_mock):
        """
        Tests the registry part of WsCapabilities
        """
        # test when registry information is read from the server (cache is outdated)
        resource = 'aresource'
        service = 'myservice'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        resource_cap_url = 'http://www.canfar.net/myservice'
        cadcreg_content = ('#test content\n {} = {} \n'
                           'ivo://some.provider/service = http://providerurl.test/service').\
            format(resource_id, resource_cap_url)
        response = Mock(text=cadcreg_content)
        get_mock.return_value = response
        # set the modified time of the cache file to 0 to make sure the info is retrieved from server
        file_modtime_mock.return_value = 0
        anon_subject = auth.Subject()
        # test anonymous access
        fh_mock = Mock()
        file_mock.write = fh_mock
        client = Mock(resource_id=resource_id)
        response = Mock()
        response.text = cadcreg_content
        client.get.return_value = response
        caps = ws.WsCapabilities(client)
        self.assertEqual(os.path.join(ws.CACHE_LOCATION, ws.REGISTRY_FILE), caps.reg_file)
        self.assertEqual(os.path.join(ws.CACHE_LOCATION, 'canfar.phys.uvic.ca', service), caps.caps_file)
        self.assertEqual(resource_cap_url, caps._get_capability_url())
        file_mock.assert_called_once_with(os.path.join(ws.CACHE_LOCATION, ws.REGISTRY_FILE), 'w')
        # TODO not sure why need to access write this way
        file_mock().__enter__.return_value.write.assert_called_once_with(cadcreg_content)

        # test when registry information is retrieved from the cache file
        get_mock.reset_mock()
        get_mock.return_value = None
        file_modtime_mock.reset_mock()
        file_mock.reset_mock()
        resource_cap_url2 = 'http://www.canfar.net/myservice2'
        cache_content2 = ('#test content\n {} = {} \n'
                           'ivo://some.provider/service = http://providerurl.test/service').\
            format(resource_id, resource_cap_url2)
        file_modtime_mock.return_value = time.time()
        file_mock().__enter__.return_value.read.return_value = cache_content2
        caps = ws.WsCapabilities(client)
        self.assertEqual(resource_cap_url2, caps._get_capability_url())

        # test when registry information is outdated but there are errors retrieving it from the CADC registry
        # so in the end go back and use the cache version
        file_modtime_mock.reset_mock()
        file_mock.reset_mock()
        file_modtime_mock.return_value = 0
        file_mock().__enter__.return_value.read.return_value = cache_content2
        get_mock.side_effect = [exceptions.HttpException()]
        client.get.side_effect = [exceptions.HttpException]
        caps = ws.WsCapabilities(client)
        self.assertEqual(resource_cap_url2, caps._get_capability_url())

    @patch('cadcutils.net.ws.os.path.getmtime')
    @patch('cadcutils.net.ws.open', mock=mock_open())
    @patch('cadcutils.net.ws.BaseWsClient.get')
    def test_get_caps(self, get_mock, file_mock, file_modtime_mock):
        """
        Tests the capabilities part of WsCapabilities
        """
        # test when registry information is read from the server (cache is outdated)
        resource = 'aresource'
        service = 'myservice'
        resource_id = 'ivo://canfar.phys.uvic.ca/{}'.format(service)
        resource_cap_url = 'www.canfar.net/myservice'
        cadcreg_content = ('#test content\n {} = {} \n'
                           'ivo://some.provider/service = http://providerurl.test/service'). \
            format(resource_id, resource_cap_url)
        # set the modified time of the cache file to 0 to make sure the info is retrieved from server
        file_modtime_mock.return_value = 0
        # test anonymous access
        fh_mock = Mock()
        file_mock.write = fh_mock
        client = Mock(resource_id=resource_id, subject=auth.Subject())
        response = Mock(text=capabilities__content.replace('WS_URL', resource_cap_url))
        client.get.return_value = response
        caps = ws.WsCapabilities(client)
        # mock _get_capability_url to return some url without attempting to access the server
        def get_url():
            return 'http://some.url/capabilities'
        caps._get_capability_url = get_url
        caps.caps_urls[service] = '{}/capabilities'.format(resource_cap_url)
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('http://{}/pub'.format(resource_cap_url),
                          caps.get_access_url('vos://cadc.nrc.ca~service/CADC/mystnd01'))
        resource_url = urlparse(resource_id)
        file_mock.assert_called_once_with(os.path.join(ws.CACHE_LOCATION,
                                                       resource_url.netloc, resource_url.path.strip('/')), 'w')
        # TODO not sure why need to access write this way
        file_mock().__enter__.return_value.write.assert_called_once_with(
            capabilities__content.replace('WS_URL', resource_cap_url))

        # repeat for basic auth subject. Mock the netrc library to prevent a lookup for $HOME/.netrc
        with patch('cadcutils.net.auth.netrclib') as netrclib_mock:
            client = Mock(resource_id=resource_id, subject=auth.Subject(netrc=True))
        client.get.return_value = response
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # capabilities works even if it has only one anonymous interface
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url),
                            caps.get_access_url('ivo://ivoa.net/std/VOSI#capabilities'))
        # same for availability
        self.assertEqual('http://{}/availability'.format(resource_cap_url),
                            caps.get_access_url('ivo://ivoa.net/std/VOSI#availability'))

        # repeat for https
        with patch('os.path.isfile') as p:
            client = Mock(resource_id=resource_id, subject=auth.Subject(certificate='somecert.pem'))
        client.get.return_value = response
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # capabilities works even if it has only one anonymous interface
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url),
                            caps.get_access_url('ivo://ivoa.net/std/VOSI#capabilities'))
        # same for availability
        self.assertEqual('http://{}/availability'.format(resource_cap_url),
                            caps.get_access_url('ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('https://{}'.format(resource_cap_url),
                          caps.get_access_url('vos://cadc.nrc.ca~service/CADC/mystnd01'))

        # test when capabilities information is retrieved from the cache file
        get_mock.reset_mock()
        get_mock.return_value = None
        file_modtime_mock.reset_mock()
        file_mock.reset_mock()
        resource_cap_url2 = 'http://www.canfar.net/myservice2'
        cache_content2 = capabilities__content.replace('WS_URL', resource_cap_url2)
        file_modtime_mock.return_value = time.time()
        file_mock().__enter__.return_value.read.return_value = cache_content2
        client = Mock(resource_id=resource_id, subject=auth.Subject())
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        caps.caps_urls[service] = '{}/capabilities'.format(resource_cap_url2)
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url2),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url2),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('http://{}/pub'.format(resource_cap_url2),
                          caps.get_access_url('vos://cadc.nrc.ca~service/CADC/mystnd01'))
        resource_url = urlparse(resource_id)
        file_mock().__enter__.return_value.read.return_value = \
            capabilities__content.replace('WS_URL', resource_cap_url2)

        # repeat for basic auth subject. Mock the netrc library to prevent a lookup for $HOME/.netrc
        with patch('cadcutils.net.auth.netrclib') as netrclib_mock:
            client = Mock(resource_id=resource_id, subject=auth.Subject(netrc=True))
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # does not work with user-password of the subject set above
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url2),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url2),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('http://{}/auth'.format(resource_cap_url2),
                          caps.get_access_url('vos://cadc.nrc.ca~service/CADC/mystnd01'))

        # repeat for https
        with patch('os.path.isfile') as p:
            client = Mock(resource_id=resource_id, subject=auth.Subject(certificate='somecert.pem'))
        caps = ws.WsCapabilities(client)
        caps._get_capability_url = get_url
        # does not work with user-password of the subject set above
        self.assertEqual('http://{}/capabilities'.format(resource_cap_url2),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#capabilities'))
        self.assertEqual('http://{}/availability'.format(resource_cap_url2),
                          caps.get_access_url('ivo://ivoa.net/std/VOSI#availability'))
        self.assertEqual('https://{}'.format(resource_cap_url2),
                          caps.get_access_url('vos://cadc.nrc.ca~service/CADC/mystnd01'))
        

# TODO By default, internet tests fail. They only succeed when test with --remote-data flag.
# Need to figure out a way to skip the tests unless that flag is present.
# class TestWsOutsideCalls(unittest.TestCase):
#     """ Class to test Ws with calls to outside sites"""
#
#     @patch('time.sleep')
#     def testCalls(self, time_mock):
#         client = ws.BaseWsClient('httpbin.org')
#         response = client.get('')
#         self.assertEqual(response.status_code, requests.codes.ok)
#
#         with self.assertRaises(requests.HTTPError):
#             client.get('status/500')
#
#         time_mock.reset_mock()
#         with self.assertRaises(requests.HTTPError):
#             client.get('status/503')
#
#         calls = [call(DEFAULT_RETRY_DELAY),
#                  call(min(DEFAULT_RETRY_DELAY*2, MAX_RETRY_DELAY)),
#                  call(min(DEFAULT_RETRY_DELAY * 4, MAX_RETRY_DELAY)),
#                  call(min(DEFAULT_RETRY_DELAY * 8, MAX_RETRY_DELAY)),
#                  call(min(DEFAULT_RETRY_DELAY * 16, MAX_RETRY_DELAY)),
#                  call(min(DEFAULT_RETRY_DELAY * 32, MAX_RETRY_DELAY))]
#
#         time_mock.assert_has_calls(calls)



# -*- coding: utf-8 -*-

# ***********************************************************************
# ******************  CANADIAN ASTRONOMY DATA CENTRE  *******************
# *************  CENTRE CANADIEN DE DONNÉES ASTRONOMIQUES  **************
#
#  (c) 2016.                            (c) 2016.
#  Government of Canada                 Gouvernement du Canada
#  National Research Council            Conseil national de recherches
#  Ottawa, Canada, K1A 0R6              Ottawa, Canada, K1A 0R6
#  All rights reserved                  Tous droits réservés
#
#  NRC disclaims any warranties,        Le CNRC dénie toute garantie
#  expressed, implied, or               énoncée, implicite ou légale,
#  statutory, of any kind with          de quelque nature que ce
#  respect to the software,             soit, concernant le logiciel,
#  including without limitation         y compris sans restriction
#  any warranty of merchantability      toute garantie de valeur
#  or fitness for a particular          marchande ou de pertinence
#  purpose. NRC shall not be            pour un usage particulier.
#  liable in any event for any          Le CNRC ne pourra en aucun cas
#  damages, whether direct or           être tenu responsable de tout
#  indirect, special or general,        dommage, direct ou indirect,
#  consequential or incidental,         particulier ou général,
#  arising from the use of the          accessoire ou fortuit, résultant
#  software.  Neither the name          de l'utilisation du logiciel. Ni
#  of the National Research             le nom du Conseil National de
#  Council of Canada nor the            Recherches du Canada ni les noms
#  names of its contributors may        de ses  participants ne peuvent
#  be used to endorse or promote        être utilisés pour approuver ou
#  products derived from this           promouvoir les produits dérivés
#  software without specific prior      de ce logiciel sans autorisation
#  written permission.                  préalable et particulière
#                                       par écrit.
#
#  This file is part of the             Ce fichier fait partie du projet
#  OpenCADC project.                    OpenCADC.
#
#  OpenCADC is free software:           OpenCADC est un logiciel libre ;
#  you can redistribute it and/or       vous pouvez le redistribuer ou le
#  modify it under the terms of         modifier suivant les termes de
#  the GNU Affero General Public        la “GNU Affero General Public
#  License as published by the          License” telle que publiée
#  Free Software Foundation,            par la Free Software Foundation
#  either version 3 of the              : soit la version 3 de cette
#  License, or (at your option)         licence, soit (à votre gré)
#  any later version.                   toute version ultérieure.
#
#  OpenCADC is distributed in the       OpenCADC est distribué
#  hope that it will be useful,         dans l’espoir qu’il vous
#  but WITHOUT ANY WARRANTY;            sera utile, mais SANS AUCUNE
#  without even the implied             GARANTIE : sans même la garantie
#  warranty of MERCHANTABILITY          implicite de COMMERCIALISABILITÉ
#  or FITNESS FOR A PARTICULAR          ni d’ADÉQUATION À UN OBJECTIF
#  PURPOSE.  See the GNU Affero         PARTICULIER. Consultez la Licence
#  General Public License for           Générale Publique GNU Affero
#  more details.                        pour plus de détails.
#
#  You should have received             Vous devriez avoir reçu une
#  a copy of the GNU Affero             copie de la Licence Générale
#  General Public License along         Publique GNU Affero avec
#  with OpenCADC.  If not, see          OpenCADC ; si ce n’est
#  <http://www.gnu.org/licenses/>.      pas le cas, consultez :
#                                       <http://www.gnu.org/licenses/>.
#
#
# ***********************************************************************

"""
Contains functionality related to interacting with Web Services. Users
of this class can instantiate the BaseWsClient in order to access
the webservices via one of the get, put, post, delete and head
functions that the service supports.

"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import logging
import sys
import time
import platform
import os

import requests
from requests import Session
from six.moves.urllib.parse import urlparse
from lxml import etree

from cadcutils import exceptions
from .. import version as cadctools_version
from . import wscapabilities

__all__ = ['BaseWsClient', 'get_resources', 'list_resources', 'DEFAULT_REGISTRY']

BUFSIZE = 8388608  # Size of read/write buffer
MAX_RETRY_DELAY = 128  # maximum delay between retries
DEFAULT_RETRY_DELAY = 30  # start delay between retries when Try_After not sent by server.
MAX_NUM_RETRIES = 6

SERVICE_RETRY = 'Retry-After'
SERVICE_AVAILABILITY_ID = 'ivo://ivoa.net/std/VOSI#availability'

# try to disable the unverified HTTPS call warnings
try:
    requests.packages.urllib3.disable_warnings()
except:
    pass


def get_resources(with_caps=True):
    """
    Fetches the registry information regarding the available resources and their
    capabilities
    :param with_caps: True for including capabilities information, false otherwise
    :return: List of the form (resource_id, capability url, wscapabilities.Capabilities)}
    """
    resources = []
    cr = wscapabilities.CapabilitiesReader()
    response = requests.get(DEFAULT_REGISTRY)
    response.raise_for_status()
    for line in response.text.split('\n'):
        caps = None
        if (not line.startswith('#')) and (len(line.strip()) > 0):
            resource_id, url = line.split('=')
            if with_caps:
                caps_resp = requests.get(url.strip())
                caps_resp.raise_for_status()
                caps = cr.parsexml(caps_resp.text)
            resources.append((resource_id.strip(), url.strip(), caps))
    return resources


def list_resources():
    """
    Displays information about the services and their capabilities.
    """
    resources = get_resources()
    for r in resources:
        if r[2] is None:
            caps_str = 'NA'
        else:
            caps_str = ', '.join(sorted(r[2]._caps.keys()))
        print('{} ({}) - Capabilities: {}\n'.format(r[0], r[1], caps_str))


class BaseWsClient(object):
    """
    Web Service client primarily for CADC services. It is a wrapper class
       around the requests module to facilitate the interaction with the
       server via the get, put, post, delete and head functions (when the service supports it).

    Additional functionality that this class transparently offer:
            - discovery the capabilities of the service and their access methods
            - handling of transient errors with retrials
            - use of the appropriate credentials when interacting with the service
            - proper agent identification for logging on the server

    Three arguments are required in order to instantiate this class:
        1. A valid resource ID corresponding to the Web service that is being accessed.
        cadcutils.net.get_resources lists all the available resources at BOOTSTRAP_LOCATION location
        2. A cadcutils.net.Subject instance that contains the user credentials (no arguments
           for anonymous subject, certificate file or basic authentication).
        3. Agent is the name and version of the application as it will be presented in the HTTP
          request e.g. foo/1.0.2
    Once the BaseWsClient has been instantiated, the get/put/post/delete/head functions can be
        called on it.

    TODO: this implementation is very tightly coupled with the requests package, especially the
        arguments of get/post/put/delete/head that are blindly passed and the response object that
        is returned and the clients work with.
       """

    def __init__(self, resource_id, subject, agent, retry=True, host=None, session_headers=None):
        """
        Client constructor
        :param resource_id -- ID of the resource being accessed (URI format) as it appears in
        the registry.
        :param subject -- The subject that is using the service
        :type subject: cadcutil.auth.Subject
        :param agent -- Name of the agent (application) that accesses the service and its version,
        e.g. foo/1.0.2
        :type agent: Subject
        :param retry -- True if the client retries on transient errors False otherwise
        :param host -- override the name of the host the service is running on (for testing purposes)
        :param session_headers -- Headers used throughout the session - dictionary format expected.
        """

        self.logger = logging.getLogger('BaseWsClient')
        logging.getLogger('BaseWsClient').addHandler(logging.NullHandler())

        if resource_id is None:
            raise ValueError('No resource ID provided')
        if agent is None or not agent:
            raise ValueError('agent is None or empty string')

        self._session = None
        self.subject = subject
        self.resource_id = resource_id
        self.retry = retry
        self.session_headers = session_headers

        # agent is / delimited key value pairs, separated by a space,
        # containing the application name and version,
        # plus the name and version of application libraries.
        # eg: foo/1.0.2 foo-lib/1.2.3
        self.agent = agent

        # Get the package name and version, plus any imported libraries.
        self.package_info = "cadcutils/{} requests/{}".format(cadctools_version.version,
                                                              requests.__version__)
        self.python_info = "{}/{}".format(platform.python_implementation(),
                                          platform.python_version())
        self.system_info = "{}/{}".format(platform.system(), platform.version())
        o_s = sys.platform
        if o_s.lower().startswith('linux'):
            distname, version, osid = platform.linux_distribution()
            self.os_info = "{} {}".format(distname, version)
        elif o_s == "darwin":
            release, version, machine = platform.mac_ver()
            self.os_info = "Mac OS X {}".format(release)
        elif o_s.lower().startswith("win32"):
            release, version, csd, ptype = platform.win32_ver()
            self.os_info = "{} {}".format(release, version)

        # build the corresponding capabilities instance
        self.caps = WsCapabilities(self, host)

        # Clients should add entries to this dict for specialized
        # conversion of HTTP error codes into particular exceptions.
        #
        # Use this form to include a search string in the response to
        # handle multiple possibilities for a single HTTP code.
        #     XXX : {'SEARCHSTRING1' : exceptionInstance1,
        #            'SEARCHSTRING2' : exceptionInstance2}
        #
        # Otherwise provide a simple HTTP code -> exception mapping
        #     XXX : exceptionInstance
        #
        # The actual conversion is performed by get_exception()
        self._HTTP_STATUS_CODE_EXCEPTIONS = {
            401: exceptions.UnauthorizedException()}

    @property
    def host(self):
        return self.caps.host

    def post(self, resource=None, **kwargs):
        """Wrapper for POST so that we use this client's session
           :param resource represents the resource to access. It can take two forms:
           1 - a URL or 2 - a tuple representing a Web Service resource in which the
           first member of the tuple is the URI of the resource (capability) and
           the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        return self._get_session().post(self._get_url(resource), **kwargs)

    def put(self, resource=None, **kwargs):
        """Wrapper for PUT so that we use this client's session
           :param resource represents the resource to access. It can take two forms:
           1 - a URL or 2 - a tuple representing a Web Service resource in which the
           first member of the tuple is the URI of the resource (capability) and
           the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        return self._get_session().put(self._get_url(resource), **kwargs)

    def get(self, resource, params=None, **kwargs):
        """Wrapper for GET so that we use this client's session
           :param resource represents the resource to access. It can take two forms:
           1 - a URL or 2 - a tuple representing a Web Service resource in which the
           first member of the tuple is the URI of the resource (capability) and
           the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        return self._get_session().get(self._get_url(resource), params=params, **kwargs)

    def delete(self, resource=None, **kwargs):
        """Wrapper for DELETE so that we use this client's session
           :param resource represents the resource to access. It can take two forms:
           1 - a URL or 2 - a tuple representing a Web Service resource in which the
           first member of the tuple is the URI of the resource (capability) and
           the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        return self._get_session().delete(self._get_url(resource), **kwargs)

    def head(self, resource=None, **kwargs):
        """Wrapper for HEAD so that we use this client's session
           :param resource represents the resource to access. It can take two forms:
           1 - a URL or 2 - a tuple representing a Web Service resource in which the
           first member of the tuple is the URI of the resource (capability) and
           the second argument is the path.
           :param kwargs additional arguments to pass to the requests.post
           :returns response as received from the request library
        """
        return self._get_session().head(self._get_url(resource), **kwargs)

    def is_available(self):
        """
        Checks whether the service is currently available or not
        :return: True if service is available, False otherwise
        """
        try:
            self.get((SERVICE_AVAILABILITY_ID, None))
        except exceptions.HttpException:
            return False
        return True

    def _get_url(self, resource):
        if type(resource) is tuple:
            # this is WS feature / path request
            path = ''
            if (resource[1] is not None) and (len(resource[1])>0):
                path = '/{}'.format(resource[1].strip('/'))
            base_url = self.caps.get_access_url(resource[0])
            access_url = '{}{}'.format(base_url, path)
            return access_url
        else:
            # assume this is url.
            resource_url = urlparse(resource)
            if resource_url.scheme not in ['http', 'https']:
                raise ValueError('Incorrect resource URL: {}'.format(resource))
            return resource

    def _get_session(self):
        # Note that the cert goes into the adapter, but we can also
        # use name/password for the auth. We may want to enforce the
        # usage of only the cert in case both name/password and cert
        # are provided.
        if self._session is None:
            self.logger.debug('Creating session.')
            self._session = RetrySession(self.retry)
            if self.subject.certificate is not None:
                self._session.cert = (self.subject.certificate, self.subject.certificate)
            else:
                if (not self.subject.anon) and (self.host is not None) and \
                        (self.subject.get_auth(self.host) is not None):
                    self._session.auth = self.subject.get_auth(self.host)

        user_agent = "{} {} {} {} ({})".format(self.agent, self.package_info, self.python_info,
                                               self.system_info, self.os_info)
        self._session.headers.update({"User-Agent": user_agent})
        if self.session_headers is not None:
            for header in self.session_headers:
                self._session.headers.update(self.session_headers)
        assert isinstance(self._session, requests.Session)
        return self._session


class RetrySession(Session):
    """ Session that automatically does a number of retries for failed transient errors. The time between retries
        double every time until a maximum of 30sec is reached

        The following network errors are considered transient:
            requests.codes.unavailable,
            requests.codes.service_unavailable,
            requests.codes.gateway_timeout,
            requests.codes.request_timeout,
            requests.codes.timeout,
            requests.codes.precondition_failed,
            requests.codes.precondition,
            requests.codes.payment_required,
            requests.codes.payment

        In addition, the Connection error 'Connection reset by remote user' also triggers a retry
        """

    retry_errors = [requests.codes.unavailable,
                    requests.codes.service_unavailable,
                    requests.codes.gateway_timeout,
                    requests.codes.request_timeout,
                    requests.codes.timeout,
                    requests.codes.precondition_failed,
                    requests.codes.precondition,
                    requests.codes.payment_required,
                    requests.codes.payment]

    def __init__(self, retry=True, start_delay=1, *args, **kwargs):
        """
        ::param retry: set to False if retries not required
        ::param start_delay: start delay interval between retries (default=1s). Note that for HTTP 503,
                             this code follows the retry timeout set by the server in Retry-After
        """
        self.logger = logging.getLogger('RetrySession')
        self.retry = retry
        self.start_delay = start_delay
        super(RetrySession, self).__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        """
        Send a given PreparedRequest, wrapping the connection to service in try/except that retries on
        Connection reset by peer.
        :param request: The prepared request to send._session
        :param kwargs: Any keywords the adaptor for the request accepts.
        :return: the response
        :rtype: requests.Response
        """
        # merge kwargs with env
        proxies = kwargs.get('proxies') or {}
        settings = self.merge_environment_settings(
            request.url, proxies, kwargs.get('stream'), kwargs.get('verify'), kwargs.get('cert'))
        kwargs.update(settings)

        if self.retry:
            current_delay = max(self.start_delay, DEFAULT_RETRY_DELAY)
            current_delay = min(current_delay, MAX_RETRY_DELAY)
            num_retries = 0
            self.logger.debug("Sending request {0}  to server.".format(request))
            current_error = None
            while num_retries < MAX_NUM_RETRIES:
                try:
                    response = super(RetrySession, self).send(request, **kwargs)
                    self.check_status(response)
                    return response
                except requests.HTTPError as e:
                    if e.response.status_code not in self.retry_errors:
                        raise exceptions.HttpException(e)
                    current_error = e
                    if e.response.status_code == requests.codes.unavailable:
                        # is there a delay from the server (Retry-After)?
                        try:
                            current_delay = int(e.response.headers.get(SERVICE_RETRY, current_delay))
                            current_delay = min(current_delay, MAX_RETRY_DELAY)
                        except Exception:
                            pass

                except requests.ConnectionError as ce:
                    current_error = ce
                    # TODO not sure this appropriate for all the 'Connection reset by peer' errors.
                    # A post/put to vospace returns a document. If operation succeeded but the error
                    # occurs during the response the code below will send the request again. Since the
                    # resource has been created/updated, a new error (bad request maybe) might be issued
                    # by the server and that can confuse the caller.
                    # This code should probably deal with HTTP errors only as the 503s above.
                    self.logger.debug("Caught exception: {0}".format(ce))
                    if ce.errno != 104:
                        # Only continue trying on a reset by peer error.
                        raise exceptions.HttpException(orig_exception=ce)
                self.logger.warning("Resending request in {}s ...".format(current_delay))
                time.sleep(current_delay)
                num_retries += 1
                current_delay = min(current_delay*2, MAX_RETRY_DELAY)
            raise exceptions.HttpException(current_error)
        else:
            response = super(RetrySession, self).send(request, **kwargs)
            self.check_status(response)
            return response

    def check_status(self, response):
        """
        Check the response status. Maps the application related requests error status into Exceptions
        and raises the others
        :param response: response
        :return:
        """
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if e.response.status_code == requests.codes.not_found:
                raise exceptions.NotFoundException(orig_exception=e)
            elif e.response.status_code == requests.codes.unauthorized:
                raise exceptions.UnauthorizedException(orig_exception=e)
            elif e.response.status_code == requests.codes.forbidden:
                raise exceptions.ForbiddenException(orig_exception=e)
            elif e.response.status_code == requests.codes.bad_request:
                raise exceptions.BadRequestException(orig_exception=e)
            elif e.response.status_code == requests.codes.conflict:
                raise exceptions.AlreadyExistsException(orig_exception=e)
            elif e.response.status_code == requests.codes.internal_server_error:
                raise exceptions.InternalServerException(orig_exception=e)
            elif e.response.status_code == requests.codes.request_entity_too_large:
                raise exceptions.ByteLimitException(orig_exception=e)
            elif self.retry and e.response.status_code in self.retry_errors:
                raise e
            else:
                raise exceptions.UnexpectedException(orig_exception=e)


DEFAULT_REGISTRY = 'http://www.cadc-ccda.hia-iha.nrc-cnrc.gc.ca/reg/resource-caps'
CACHE_REFRESH_INTERVAL = 10*60
CACHE_LOCATION = os.path.join(os.path.expanduser("~"), '.config', 'cadc-registry')
REGISTRY_FILE = 'resource-caps'


class WsCapabilities(object):
    """
    Contains the capabilities of Web Services. The most useful function is get_access_url that
    returns the url corresponding to a feature of a Web Service
    """

    def __init__(self, ws_client, host=None):
        """
        :param ws_client: WebService client that the capabilities are required for
        :param host: use this host rather than the default CADC host for reg lookup
        """
        self.logger = logging.getLogger('WsCapabilities')
        self.ws = ws_client
        self._host = host
        cache_location = CACHE_LOCATION
        if host is not None:
            cache_location = os.path.join(cache_location, 'alt-domains', host)

        if not os.path.isdir(cache_location):
            os.makedirs(cache_location)

        # check the registry file in cache if it requires a refresh
        self.reg_file = os.path.join(cache_location, REGISTRY_FILE)
        resource_id = urlparse(ws_client.resource_id)
        self.caps_file = os.path.join(cache_location, resource_id.netloc, resource_id.path.strip('/'))
        self.last_regtime = 0
        self.last_capstime = 0
        self._caps_reader = wscapabilities.CapabilitiesReader()
        self.caps_urls = {}
        self.features = {}
        self.capabilities = {}

    def get_access_url(self, feature):
        """
        Returns the access URL corresponding to a feature and the authentication information associated
        with the subject that created the Web Service client
        :param feature: Web Service feature
        :return: corresponding access URL
        """
        
        if (time.time() - self.last_capstime) > CACHE_REFRESH_INTERVAL:
            if self.last_capstime == 0:
                # startup
                try:
                    self.last_capstime = os.path.getmtime(self.caps_file)
                except OSError:
                    # cannot read the cache file for whatever reason
                    pass
            caps = self._get_content(self.caps_file, self._get_capability_url(), self.last_capstime)
            # caps is a string but it's xml content claims it's utf-8 encode, hence need to encode it before
            # parsing it.
            self.capabilities = self._caps_reader.parsexml(caps.encode('utf-8'))
            if (time.time() - self.last_capstime) > CACHE_REFRESH_INTERVAL:
                self.last_capstime = time.time()
        sms = self.ws.subject.get_security_methods()
        
        return self.capabilities.get_access_url(feature, sms)

    @property
    def host(self):
        return self._host

    def _get_content(self, resource_file, url, last_accessed):
        """
         Return content from a local cache file if information is recent (it was accessed
         less than CACHE_REFRESH_INTERVAL seconds ago). If not, it updates the cache
         from the provided url before returning the content.
        """
        content = None
        if (time.time() - last_accessed) < CACHE_REFRESH_INTERVAL:
            # get reg information from the cached file
            self.logger.debug('Read cached content of {}'.format(resource_file))
            try:
                with open(resource_file, 'r') as f:
                    content = f.read()
            except Exception:
                # will download it
                pass
        # config dirs if they don't exist yet
        if not os.path.exists(os.path.dirname(resource_file)):
            os.makedirs(os.path.dirname(resource_file))

        if content is None:
            # get information from the bootstrap registry
            try:
                content = self.ws.get(url).text
                with open(resource_file, 'w') as f:
                    f.write(content)
            except exceptions.HttpException:
                # problems with the bootstrap registry. Try to use the old local one
                # regardless of how old it is
                with open(resource_file, 'r') as f:
                    content = f.read()
        if content is None:
            raise RuntimeError("Cannot get the registry info from either local or remote source")
        return content

    def _get_capability_url(self):
        """
        Parses the registry information and returns the url of the capabilities feature
        of the Web Service
        :return: URL to the capabilities feature
        """
        if (time.time() - self.last_regtime) > CACHE_REFRESH_INTERVAL:
            if self.last_regtime == 0:
                # startup
                try:
                    self.last_regtime = os.path.getmtime(self.reg_file)
                except OSError:
                    # cannot read the cache file for whatever reason
                    pass
            
            # replace registry host name if necessary
            registry_url = DEFAULT_REGISTRY
            url = urlparse(registry_url)
            if self._host != None and url.netloc != self._host:
                registry_url = '{}://{}{}'.format(url.scheme, self._host, url.path)
            self.logger.debug('Resolved URL: {}'.format(registry_url))
            
            reg = self._get_content(self.reg_file, registry_url, self.last_regtime)
            self.caps_urls = {}
            if (time.time() - self.last_regtime) > CACHE_REFRESH_INTERVAL:
                self.last_regtime = time.time()
            # parse it
            for line in reg.split('\n'):
                if not line.startswith('#') and (len(line) > 0):
                    feature, url = line.split('=')
                    self.caps_urls[feature.strip()] = url.strip()
        if self.ws.resource_id not in self.caps_urls:
            raise AttributeError('Resource ID {} not found. Available resource IDs: {}'.
                                               format(self.ws.resource_id, self.caps_urls.keys()))
        return self.caps_urls[self.ws.resource_id]



import errno
import logging
import os
import sys
from six.moves import configparser
from shutil import copyfile


logger = logging.getLogger('config')
logger.setLevel(logging.INFO)

if sys.version_info[1] > 6:
    logger.addHandler(logging.NullHandler())


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


class Config(object):

    def __init__(self, config_path, default_config_path=None):
        logger.info("Using config file {0}.".format(config_path))

        # check config file exists and can be read
        if not os.path.isfile(config_path) and not os.access(config_path, os.R_OK):
            error = "Can not read {0}.".format(config_path)
            logger.error(error)
            raise IOError(error)

        self.parser = configparser.ConfigParser()
        if default_config_path:
            try:
                self.parser.readfp(open(default_config_path))
            except configparser.Error as exc:
                logger.error("Error opening {0} because {1}.".format(default_config_path, exc.message))

        try:
            self.parser.read(config_path)
        except configparser.Error as exc:
            logger.error("Error opening {0} because {1}.".format(config_path, exc.message))

    def get(self, section, option):
        try:
            return self.parser.get(section, option)
        except configparser.NoOptionError:
            pass
        return None

    @staticmethod
    def write_config(config_path, default_config_path):
        """
        :param config_path:
        :param default_config_path:
        :return:
        """

        # if not local config file then write the default config file
        if not os.path.isfile(config_path):
            mkdir_p(os.path.dirname(config_path))
            copyfile(default_config_path, config_path)
            return

        # read local config file
        parser = configparser.ConfigParser()
        try:
            parser.read(config_path)
        except configparser.Error as exc:
            logger.error("Error opening {0} because {1}.".format(config_path, exc.message))
            return

        # read default config file
        default_parser = configparser.RawConfigParser()
        try:
            default_parser.read(default_config_path)
        except configparser.Error as exc:
            logger.error("Error opening {0} because {1}.".format(default_config_path, exc.message))
            return

        # update config file with new options from the default config
        updated = False
        for section in default_parser.sections():
            default_items = default_parser.items(section)
            for option, value in default_items:
                if not parser.has_section(section):
                    parser.add_section(section)
                if not parser.has_option(section, option):
                    parser.set(section, option, value)
                    updated = True

        # remove old options not in the default config file?
        # for section in default_parser.sections():
        #     options = parser.options(section)
        #     for option in options:
        #         if not default_parser.has_option(section, option):
        #             parser.remove_option(section, option)

        # write updated config file
        if updated:
            try:
                config_file = open(config_path, 'w')
                parser.write(config_file)
                config_file.close()
            except Exception as exc:
                print (exc.message)
